#    Diagnostic and test task for the AGATA Digitizer interface

set frame  Agata
set title "Agata Digitizer Access Client"
set icon   Agata

set started 0

set LogFlag 1

set inputcolour  green4
set outputcolour red
set insertcolour blue
set bkgcolour    azure

set inputprompt  ""
set outputprompt ""
set insertprompt ">>"

set AgataServerList {}
set AgataServer "select a server"
variable WebServer

set ResourceList { \
       "Core: Virtex for Segment ADC card 1" \
       "Core: Virtex for Segment ADC card 2" \
       "Core: Virtex for core ADCs" \
       "Core: Main Board" \
       "Segment: Virtex for Segment ADC card 1" \
       "Segment: Virtex for Segment ADC card 2" \
       "Segment: Virtex for Segment ADC card 3" \
       "Segment: Virtex for Segment ADC card 4" \
       "Segment: Main Board" \
}
set ResourceIDs {0 1 2 3 8 9 10 11 12}


set FpgaList {}
set FpgaIDs {}
set Fpga ""

set AddressModifierList {A16 A32}
set AddressModifier [lindex $AddressModifierList 0]

set NumDigitizer 0
set DigitizerList {}

set Digitizer -1         ;#   currently selected digitizer
set CorePresent 0        ;#   for current digitizer
set SegmentPresent 0     ;#   for current digitizer
for {set i 0} {$i <= 12} {incr i} {set Present($i) 0}     ;# index is fpga ID


set Run 0    ;#  temperature monitor

set SegmentTempList { \
       "Segment: Virtex for Segment ADC card 1" \
       "Segment: Analog for Segment ADC card 1" \
       "Segment: Virtex for Segment ADC card 2" \
       "Segment: Analog for Segment ADC card 2" \
       "Segment: Virtex for Segment ADC card 3" \
       "Segment: Analog for Segment ADC card 3" \
       "Segment: Virtex for Segment ADC card 4" \
       "Segment: Analog for Segment ADC card 4" \
       "Segment: PSU #1" \
       "Segment: PSU #2" \
}

set CoreTempList { \
       "Core: Virtex for Segment ADC card 1" \
       "Core: Analog for Segment ADC card 1" \
       "Core: Virtex for Segment ADC card 2" \
       "Core: Analog for Segment ADC card 2" \
       "Core: Virtex for core ADCs" \
       "Core: Analog for core ADCs" \
       "Core: PSU #1" \
       "Core: PSU #2" \
       "Core: PSU #3" \
}


set Address 0x0
set Data 0x0
set Length 1
set BlkData 0x0000

set ShortTimeOut 5000
set verbose 0


proc bgerror err {
    global errorInfo tkerrorInfo tkskiperror
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    if {$tkskiperror == 1} then {set tkskiperror 0; return}
    set tkerrorInfo $errorInfo

    if {[grab current .] != ""} then {grab release [grab current .]}

    catch {:modify [self frame-name] :busy T}
    if {![info exists FaultCode] || $FaultCode == "NONE"} {
        set z [catch {midas_dialog [self frame-name].tkerrorDialog "Error Detected in Tcl Script" \
                "Task '[Get_My_Name]' detected error in Tcl script\n\n$err" "error" 0 "continue" "Stack Trace"} button]
    } else {
        set z [catch {midas_dialog [self frame-name].tkerrorDialog "Error Detected in Tcl Script" \
                "Task '[Get_My_Name]' detected error in Tcl script\n\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo" \
                       "error" 0 "continue"} button]
    }
    catch {:modify [self frame-name] :busy F}

    if {$z != 0} then {
       puts stderr "midas_dialog: [self frame-name].tkerrorDialog: $button"
       destroy [self frame-name].tkerrorDialog
       return
    }

    if {$button != 0} then {tkStackTrace}
}



proc make-frame {} {
   global env
   global AddressModifierList Address Data Length BlkData

set rows 16
set cols 80

  :frame [self frame-name] :label "[self frame-title]" \
     :icon-name "[self icon-title]"

  :panel P0 :layout V :padx 5 :pady 5
    :layout H; :col-gap 0
    :text XPORTADDRT         :label "Agata Digitizer Web Service" :w 20 :value "select a server"
    :choice-stack XPORTADDRC :label "" :strings "" 
    :hskip 70
    :checkbox VERBOSE        :label "Enable Verbose Option"
 
  :panel P3 :layout V :padx 5 :pady 5
      :layout H; :row-gap 10; :col-gap 30
      :choice-stack DIGITIZER :label "Digitizer"  :exclusive T :menu "" :value -1
      :choice-stack FPGA      :label "FPGA"  :exclusive T :strings "" :value -1
      :hskip 120
      :choice-stack ADDRMOD   :label "Address Modifier"  :exclusive T :strings $AddressModifierList :value 0
      :next-row T
      :text ADDR              :label "Address/Cmd" :w 10 :value $Address
      :text DATA     :label "Set Data"    :w 10 :value $Data
      :text LENGTH   :label "Read Block Length" :w 6 :value $Length
      :next-row T
      :text BLKDATA  :label "Write Block Data"  :w 80 :value $BlkData
      :next-row T
      :button GET    :label "Do Get"  :w 14
      :button SET    :label "Do Set"  :w 14
      :button READ   :label "Do Read Block"  :w 14
      :button WRITE  :label "Do Write Block" :w 14

  :panel P2 :layout V :padx 5 :pady 5
    :galley LOG :rows $rows :cols $cols :font fixed
     configuretags

  :panel P1 :layout V :padx 5 :pady 5
    :layout H; :col-gap 0
     :button CLEAR   :label "Clear log window" :command ":modify LOG :empty T"
     :button TEMP    :label "Temperature Monitor" :command Monitor_Temperature

   open-frame

  :galley_size LOG $rows $cols

#    focus [:path CMD].entry

   :show T
}


proc click-configure {W} {:galley_resize LOG}

proc configuretags {} {
   global inputcolour outputcolour insertcolour bkgcolour

    :modify LOG :createtag "taginput  -foreground $inputcolour  -background $bkgcolour"
    :modify LOG :createtag "tagoutput -foreground $outputcolour -background $bkgcolour"
    :modify LOG :createtag "taginsert -foreground $insertcolour -background $bkgcolour"
}

proc insert {s} {:modify LOG :usingtag taginsert :strings $s\n}

proc start args {
   global started env Options AgataServerList

   if {$started == 0} then {make-frame} else {open-frame}
   set started 1

#    load application options and establish default values for any not defined

   load_Options

   if {![info exists Options(WebServer)]}     {set Options(WebServer) "localhost"}

   update_Options

   foreach option [list WebServer] {
       global $option
       set $option $Options($option)
   }

   set  AgataServerList [list $WebServer]    ;# put default into the menu
   :modify XPORTADDRC :strings $AgataServerList :value [lsearch $AgataServerList $WebServer]
   
}

proc Save_Options {} {
   global Options

   foreach option [list WebServer] {
       global $option
       set Options($option) [set $option]
   }

   update_Options
}

proc click arg {
   global history verbose longbuffer datafile usedatafile
   set w [lindex $arg 2]
   set v [lindex $arg 3]
   set i [lindex $arg 4]

   clear-footer
   switch -glob $w  {

      XPORTADDRT    {click-server $v}
      XPORTADDRC    {click-server $i}
      
      VERBOSE       {set verbose $v; AGATADigitizerAccessClient__AccessServerVerbose $verbose}

      FPGA          {click-fpga $i}
 
      default {
         set p "click-[lowercase $w]"
         if {[info procs $p] == "$p"} then {$p $v} else {set-footer "no handler for widget $w"}
      }
   }
}

proc click-server {v} {
    global env AgataServerList WebServer AGATADigitizerAccessServerHost
    global NumDigitizer DigitizerList Digitizer
    global verbose

    set WebServer $v
    set AgataServerList [list $WebServer]    ;# put default into the menu
    :modify XPORTADDRT :value $WebServer
    :modify XPORTADDRC :strings $AgataServerList :value [lsearch $AgataServerList $WebServer]

    Save_Options

    set AGATADigitizerAccessServerHost $WebServer
    package require AGATADigitizerAccessClient

    loadsource AGATADigitizerAccessClient     ;# better solution needed

    AGATADigitizerAccessClient__AccessServerVerbose $verbose
    set NumDigitizer [AGATADigitizerAccessClient__Init]

    insert "connected to Agata Digitizer Web Service at host $WebServer: $NumDigitizer digitizer(s) discovered"

    set DigitizerList {}
    for {set i 0} {$i < $NumDigitizer} {incr i} {lappend DigitizerList $i}
    :modify DIGITIZER :menu $DigitizerList :value -1
    
    if {[llength $DigitizerList] == 1} {click-digitizer [first $DigitizerList]} else {insert "select a digitizer"}
}

proc click-data {v} {
    global Data

    set Data $v

    insert "Data= [format "0x%x" $Data]"
}

proc click-blkdata {v} {
    global BlkData

    set BlkData $v
}

proc click-length {v} {
    global Length

    set Length $v

    insert "Read Block Length= [format "0x%x" $Length]"
}


proc click-addr {v} {
    global Address

    set Address $v

    insert "Address/Cmd= [format "0x%x" $Address]"
}

proc click-digitizer {v} {
    global Digitizer Present CorePresent SegmentPresent
    global FpgaList FpgaIDs
    global ShortTimeOut

    set Digitizer $v
    set CorePresent 0
    set SegmentPresent 0
    for {set i 0} {$i <= 12} {incr i} {set Present($i) 0} 

    insert "Digitizer= $Digitizer"

    set FpgaList {}
    set FpgaIDs {}

    insert "probing for Core Main Board"

    set fpga 3     ;# core main board
    set cmd 14     ;# safe read command for test
    set Length 6   ;# returns 6 bytes
    set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
    set rc [catch {AGATADigitizerAccessClient__Read $Digitizer $fpga $cmd A16 $Length} m]
    if {$rc == 0} {
        insert "found it"
        lappend FpgaList "Core: Main Board"
        lappend FpgaIDs $fpga
        set CorePresent 1
        set Present($fpga) 1
    } else {
        insert "Error returned [first [split $m \n]]"
     }

    if {$CorePresent == 1} {

       insert "probing for Core ADCs"

       set fpga 2    ;# core ADCs
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Core: Virtex for core ADCs"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }

       insert "probing for Core segment ADC card #1"

       set fpga 0    ;# core segment ADC card #1
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Core: Virtex for Segment ADC card 1"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }

       insert "probing for Core segment ADC card #2"

       set fpga 1    ;# core segment ADC card #2
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Core: Virtex for Segment ADC card 2"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }
    }

    insert "probing for Segment Main Board"

    set fpga 12    ;# segment main board
    set cmd 14     ;# safe read command for test
    set Length 6   ;# returns 6 bytes
    set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
    set rc [catch {AGATADigitizerAccessClient__Read $Digitizer $fpga $cmd A16 $Length} m]
    if {$rc == 0} {
        insert "found it"
        lappend FpgaList "Segment: Main Board"
        lappend FpgaIDs $fpga
        set SegmentPresent 1
        set Present($fpga) 1
    } else {
        insert "Error returned [first [split $m \n]]"
    }

    if {$SegmentPresent == 1} {

       insert "probing for Segment ADC card #1"

       set fpga 8    ;# Segment ADC card #1
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 1"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }

       insert "probing for Segment ADC card #2"

       set fpga 9    ;# Segment ADC card #2
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 2"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }

       insert "probing for Segment ADC card #3"

       set fpga 10    ;# Segment ADC card #3
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 3"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }

       insert "probing for Segment ADC card #4"

       set fpga 11    ;# Segment ADC card #4
       set address 15
       set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer  $ShortTimeOut} m]
       set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $address A16} m]
       if {$rc == 0} {
           insert "found it"
           lappend FpgaList "Segment: Virtex for Segment ADC card 4"
           lappend FpgaIDs $fpga
           set Present($fpga) 1
       } else {
           insert "Error returned [first [split $m \n]]"
       }
   }

   set rc [catch {AGATADigitizerAccessClient__SetTimeOut $Digitizer 0} m]   ;# reset timeout to default

   if {[expr $CorePresent + $SegmentPresent] == 0} {
       insert "Digitizer not found"
       :modify DIGITIZER  :value -1
       :modify FPGA :strings "" :value -1
   } else {
       :modify FPGA :strings $FpgaList :value 0
       click-fpga [first $FpgaList]
   }

}


proc click-fpga {v} {
    global Fpga

    set Fpga $v

    insert "Fpga= $Fpga"
}

proc click-addrmod {v} {
    global AddressModifierList AddressModifier

    set AddressModifier [lindex $AddressModifierList $v]

    insert "AddressModifier= $AddressModifier"
}

proc click-set {args} {
    global Digitizer FpgaList FpgaIDs Fpga Address AddressModifier Data
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

    set rc [catch {AGATADigitizerAccessClient__Set $Digitizer $fpga $Address $AddressModifier $Data} m]
    if {$rc != 0} {
        insert "Set returned with an error\n$m\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo"
    } else {
        insert "Set returned $m ([format "0x%x" $m])"
    }
}

proc click-get {args} {
    global Digitizer FpgaList FpgaIDs Fpga Address AddressModifier
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

    set rc [catch {AGATADigitizerAccessClient__Get $Digitizer $fpga $Address $AddressModifier} m]
    if {$rc != 0} {
        insert "Get returned with an error\n$m\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo"
    } else {
        insert "Get returned $m ([format "0x%x" $m])"
    }
}

proc click-read {args} {
    global Digitizer FpgaList FpgaIDs Fpga Address Length AddressModifier
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

    set rc [catch {AGATADigitizerAccessClient__Read $Digitizer $fpga $Address $AddressModifier $Length} m]
    if {$rc != 0} {
        insert "Read returned with an error\n$m\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo"
    } else {
        insert "Read returned $m"
    }
}

proc click-write {args} {
    global Digitizer FpgaList FpgaIDs Fpga Address Length AddressModifier BlkData
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    set fpga [lindex $FpgaIDs [lsearch $FpgaList $Fpga]]

    set rc [catch {AGATADigitizerAccessClient__Write $Digitizer $fpga $Address $AddressModifier [llength $BlkData] $BlkData} m]
    if {$rc != 0} {
        insert "Write returned with an error\n$m\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo"
    } else {
        insert "Write returned $m"
    }
}


#    The Options are experiment specific and normally used by SYS
#    However the same code as for the session preferences is used which allows for wider use in the future.

proc get_Option_file {} {
    global env

#    look for the Option file
    set pd $env(EXPERIMENTBASE)
    if {![file exists $pd]} then {
         insert-log "$pd does not exist - Join an experiment!"
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd MidasOptions]
    set z [catch {set f [open $pf r]} m]
    if {$z != 0} then {
        set z [catch {set f [open $pf w]} m]
        if {$z != 0} then {
            insert-log "Unable to access $pf - $m"
            return 1
        }
        close $f
        return {}
    }

#    and read it
    set lines [split [read $f] "\n"]
    close $f

    return $lines
}

proc put_Option_file {lines} {
   global env

#    look for the Option file
    set pd $env(EXPERIMENTBASE)
    if {![file exists $pd]} then {
         insert-log "$pd does not exist - Join an experiment!"
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd MidasOptions]
    set z [catch {set f [open $pf w]} m]
    if {$z != 0} then {
        insert-log "Unable to access $pf - $m"
        return 1
    }

    foreach item $lines {
      if {$item != ""} {puts $f $item}
    }

    close $f
    return 0
}

#    save all Options for the current task
proc update_Options {} {
   global env Options

   set lines [get_Option_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]
   if {[llength $range] > 0} then {
       set lines [lreplace $lines [first $range] [expr [second $range] + 1]]
   }

   lappend lines "## $key"
   foreach item [array names Options] {
       lappend lines "$item $Options($item)"
   }
   lappend lines "##"

   set z [put_Option_file $lines]
   return $z
}

#    load all Options for the current task
proc load_Options {} {
   global env Options

   set lines [get_Option_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           set Options([head $item]) [tail $item]
       }
   }
}

#    return task Options
proc get_Options {task} {
   global env

   set lines [get_Option_file]
   set key $task
   set range [find-keyword $lines $key]

   set Options {}

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           lappend Options [list [first $item] [second $item]]
       }
   } else {
       lappend Options [list Options.filename  "[file join $env(HOME) MidasOptions Options]"]
   }
   return $Options
}


proc Monitor_Temperature {} {
   global env
   global ResourceList

  :frame [self frame-name].temp :label "[self frame-title]: Temperature Monitor" :icon-name "[self icon-title]"

  for {set i 0} {$i < 3} {incr i} {

      :panel P$i :layout V :padx 0 :pady 0
        :layout H

      for {set j 0} {$j < 3} {incr j} {

         set c [expr ($i*3)+$j]
         set resource [lindex $ResourceList $c]
         :canvas C$c :w 400 :h 300 
            [:path C$c] configure -background black
      }
  }

    :panel LAST :layout V :padx 5 :pady 5
    :layout H; :col-gap 20

      :button DONE   :label "Done"  :command "frame-quit [self frame-name].temp"
      :button TSTART :label "Start" :command Start_TempMon
      :button TSTOP  :label "Stop"  :command Stop_TempMon
      :button RESET  :label "Reset" :command Reset_TempMon

    :show T

   Reset_TempMon
   Stop_TempMon

}

proc Reset_TempMon {} {
    global CoreTemperature SegmentTemperature

    for {set i 0} {$i <= 8} {incr i} {
        set CoreTemperature($i) {}
        for {set j 0} {$j < 400} {incr j} {lappend CoreTemperature($i) 20.0}
    }

    for {set i 0} {$i <= 9} {incr i} {
        set SegmentTemperature($i) {}
        for {set j 0} {$j < 400} {incr j} {lappend SegmentTemperature($i) 20.0}
    }

}

proc Start_TempMon {} {
    global Run
 
    set Run 1
    Update_Temperature
}

proc Stop_TempMon {} {
    global Run
 
    set Run 0
}

proc Update_Temperature {} {
    global Digitizer
    global CorePresent CoreTemperature
    global SegmentPresent SegmentTemperature
    global Run

    set-footer "Updating temperatures..."

    set Error 0

    if {$CorePresent == 1} {

       set fpga 3         ;# core main board
       set cmd 19         ;# use command 19
       set Length 20      ;# which returns 20 bytes of data
       set rc [catch {AGATADigitizerAccessClient__Read $Digitizer $fpga $cmd A16 $Length} m]
       if {$rc == 0} {
            insert "$m"
            set T [format_readings $m]
            for {set i 0} {$i <= 8} {incr i} {
                set CoreTemperature($i) [linsert [lreplace $CoreTemperature($i) end end] 0 [lindex $T $i]]
            }
       } else {
           insert "Error returned $m"
           incr Error
       }
    }

    if {$SegmentPresent == 1} {

       set fpga 12        ;# segment main board
       set cmd 19         ;# use command 19
       set Length 20      ;# which returns 20 bytes of data
       set rc [catch {AGATADigitizerAccessClient__Read $Digitizer $fpga $cmd A16 $Length} m]
       if {$rc == 0} {
            insert "$m"
            set T [format_readings $m]
            for {set i 0} {$i <= 9} {incr i} {
                set SegmentTemperature($i) [linsert [lreplace $SegmentTemperature($i) end end] 0 [lindex $T $i]]
            }
       } else {
           insert "Error returned $m"
           incr Error
       }
    }

    clear-footer

    if {$Error == 0 && $Run == 1} {after 5000 Update_Temperature}

    Update_Graph

}

proc format_readings {Temps} {

    set T {}
    for {set i 0; set j 0} {$i < 20} {incr i; incr j} {
        set msb [lindex $Temps $i]
        incr i
        set lsb [lindex $Temps $i]
        set t [expr (((($msb * 255) + $lsb)/8) & 0x0fff) * 0.0625]
        if {$msb >= 128} then {
           lappend T [expr $t * -1.0]
        } else {
           lappend T $t 
        }
    }
insert "[lindex $T 4] [lindex $T 5] [lindex $T 6] [lindex $T 7] [lindex $T 8]"
    return $T
}

proc Update_Graph {} {
    global Digitizer
    global CoreTemperature SegmentTemperature
    global Present

    show_reset

#    do Core

    if {$Present(0) == 1} {            ;# Segment ADC board 1 present
        show_core 0                     ;#  virtex
        show_core 1                     ;#  analog
    }
    if {$Present(1) == 1} {            ;# Segment ADC board 1 present
        show_core 2                     ;#  virtex
        show_core 3                     ;#  analog
    }
    if {$Present(2) == 1} {           ;# Core ADC board present 
        show_core 4                     ;#  virtex
        show_core 5                     ;#  analog
    }
    if {$Present(3) == 1} {           ;# core board 
        show_core 6                     ;#  psu #1
        show_core 7                     ;#  psu #2
        show_core 8                     ;#  psu #3
    }

#    do Segment

    if {$Present(8) == 1} {            ;# Segment ADC board 1 present 
        show_segment 0                     ;#  virtex
        show_segment 1                     ;#  analog
    }
    if {$Present(9) == 1} {            ;# Segment ADC board 2 present 
        show_segment 2                     ;#  virtex
        show_segment 3                     ;#  analog
    }
    if {$Present(10) == 1} {           ;# Segment ADC board 3 present 
        show_segment 4                     ;#  virtex
        show_segment 5                     ;#  analog
    }
    if {$Present(11) == 1} {           ;# Segment ADC board 4 present 
        show_segment 6                     ;#  virtex
        show_segment 7                     ;#  analog
    }
    if {$Present(12) == 1} {            ;# Core board present 
        show_segment 8                     ;#  psu #1
        show_segment 9                     ;#  psu #2
    }

}

proc show_reset {} {
    global Canvas

    for {set i 0} {$i < 9} {incr i} {
        set _canvas [self frame-name].temp.p[expr $i/3].c$i
        :draw modify :tag t$i :delete T
        :draw modify :tag l$i :delete T
    }

    set Canvas 0
}

proc show_segment {n} {
    global SegmentTemperature SegmentTempList
    global _canvas
    global Canvas

    set _canvas [self frame-name].temp.p[expr $Canvas/3].c$Canvas

    :draw modify :tag t$n :delete T
    :draw modify :tag l$n :delete T

    for {set i 0; set j 1} {$i < 399} {incr i; incr j} {
        set t0 [lindex $SegmentTemperature($n) $i]
        set t0 [expr int (300 - (($t0 - 20) * 5))]
        set t1 [lindex $SegmentTemperature($n) $j]
        set t1 [expr int (300 - (($t1 - 20) * 5))]
        :draw line :tag l$n :colour red :value $i $t0 $j $t1
    }

    :draw text :tag t$n :colour white :text "[lindex $SegmentTempList $n]" :value 5 5

    incr Canvas
}

proc show_core {n} {
    global CoreTemperature CoreTempList
    global _canvas
    global Canvas

    set _canvas [self frame-name].temp.p[expr $Canvas/3].c$Canvas

    :draw modify :tag t$n :delete T
    :draw modify :tag l$n :delete T

    for {set i 0; set j 1} {$i < 399} {incr i; incr j} {
        set t0 [lindex $CoreTemperature($n) $i]
        set t0 [expr int (300 - (($t0 - 20) * 5))]
        set t1 [lindex $CoreTemperature($n) $j]
        set t1 [expr int (300 - (($t1 - 20) * 5))]
        :draw line :tag l$n :colour yellow :value $i $t0 $j $t1
    }

    :draw text :tag t$n :colour white :text "[lindex $CoreTempList $n]" :value 5 5

    incr Canvas
}
